from geojson import Feature

from flatgeobuf.packedrtree import Rect

try:
    import shapely.geometry
    import shapely.prepared
except ImportError:
    shapely = None


# NOTE: Assuming GeoJSON
class BBoxFilter:
    def __init__(self, bbox: Rect):
        if shapely is None:
            return

        prep_bbox = shapely.prepared.prep(shapely.geometry.box(*bbox, ccw=True))
        self.prep_bbox = prep_bbox

    def has_intersection(self, feature: Feature) -> bool:
        """Check if a feature intersects with a bbox."""

        # NOTE: If shapely is not installed, we assume that the feature intersects
        if shapely is None:
            return True

        shape = shapely.geometry.shape(feature.geometry)

        return self.prep_bbox.intersects(shape)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Callable, List, Optional, TypedDict, Union, cast

from flatgeobuf.FlatGeobuf.Geometry import Geometry
from flatgeobuf.FlatGeobuf.GeometryType import GeometryType


class ParsedGeometry(TypedDict, total=False):
    xy: List[float] | None
    z: List[float] | None
    ends: List[int] | None
    parts: List["ParsedGeometry"] | None
    type: GeometryType | None


class SimpleGeometry(metaclass=ABCMeta):
    @abstractmethod
    def get_flat_coordinates(self) -> List[float]:
        pass

    @abstractmethod
    def get_type(self) -> str:
        pass


class BasePolygon(SimpleGeometry):
    @abstractmethod
    def get_ends(self) -> List[int]:
        pass


class BaseMultiLineString(SimpleGeometry):
    @abstractmethod
    def get_ends(self) -> List[int]:
        pass


class BaseMultiPolygon(SimpleGeometry):
    @abstractmethod
    def get_endss(self) -> List[List[int]]:
        pass

    @abstractmethod
    def get_polygons(self) -> List[BasePolygon]:
        pass


CreateGeometryCallable = Callable[
    [Union[Geometry, None], GeometryType], Optional[SimpleGeometry]
]


def build_geometry(parsed_geometry: ParsedGeometry) -> Geometry:
    raise NotImplementedError()
    # xy = parsed_geometry.get("xy", None)
    # z = parsed_geometry.get("z", None)
    # ends = parsed_geometry.get("ends", None)
    # parts = parsed_geometry.get("parts", None)
    # type = parsed_geometry.get("type", None)

    # if parts:
    #     part_offsets = [build_geometry(part) for part in parts]
    #     parts_offset = Geometry.create_parts_vector(part_offsets)
    #     geometry = Geometry(xy=[], z=[], ends=[], parts=parts_offset, type=type)
    #     return geometry

    # xy_offset = Geometry.create_xy_vector(xy)
    # z_offset = Geometry.create_z_vector(z) if z else None
    # ends_offset = Geometry.create_ends_vector(ends) if ends else None

    # geometry = Geometry(xy=xy_offset, z=z_offset, ends=ends_offset, type=type)
    # return geometry


def flat(
    a: List[Union[List[float], float]], xy: List[float], z: List[float]
) -> Optional[List[float]]:
    if len(a) == 0:
        return None

    if isinstance(a[0], list):
        for sa in a:
            flat(sa, xy, z)
    else:
        if len(a) == 2:
            xy.extend(a)
        else:
            xy.extend(a[:2])
            z.append(a[2])


def parse_geometry(
    geometry: SimpleGeometry, header_geom_type: GeometryType
) -> ParsedGeometry:
    xy = None
    ends = None
    parts = None

    type = header_geom_type
    if type == GeometryType.Unknown:
        type = to_geometry_type(geometry.get_type())

    if type == GeometryType.MultiLineString:
        if hasattr(geometry, "get_flat_coordinates"):
            xy = geometry.get_flat_coordinates()
        mls_ends = cast(BaseMultiLineString, geometry).get_ends()
        if len(mls_ends) > 1:
            ends = [e >> 1 for e in mls_ends]
    elif type == GeometryType.Polygon:
        if hasattr(geometry, "get_flat_coordinates"):
            xy = geometry.get_flat_coordinates()
        p_ends = cast(BasePolygon, geometry).get_ends()
        if len(p_ends) > 1:
            ends = [e >> 1 for e in p_ends]
    elif type == GeometryType.MultiPolygon:
        mp = cast(BaseMultiPolygon, geometry)
        parts = [parse_geometry(p, GeometryType.Polygon) for p in mp.get_polygons()]
    else:
        if hasattr(geometry, "get_flat_coordinates"):
            xy = geometry.get_flat_coordinates()

    parsed_geometry: ParsedGeometry = {
        "xy": xy,
        "ends": ends,
        "type": type,
        "parts": parts,
    }

    return parsed_geometry


def pair_flat_coordinates(
    xy: List[float], z: Optional[List[float]] = None
) -> List[List[float]]:
    new_array = []
    for i in range(0, len(xy), 2):
        a = [xy[i], xy[i + 1]]
        if z:
            a.append(z[i >> 1])
        new_array.append(a)
    return new_array


def to_geometry_type(name: Optional[str]) -> GeometryType:
    if not name:
        return GeometryType.Unknown
    return getattr(GeometryType, name)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FlatGeobuf

import flatbuffers
from flatbuffers.compat import import_numpy
from typing import Any
from typing import Optional
np = import_numpy()

class Column(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset: int = 0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Column()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsColumn(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Column
    def Init(self, buf: bytes, pos: int):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Column
    def Name(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Column
    def Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Column
    def Title(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Column
    def Description(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Column
    def Width(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return -1

    # Column
    def Precision(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return -1

    # Column
    def Scale(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return -1

    # Column
    def Nullable(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return True

    # Column
    def Unique(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Column
    def PrimaryKey(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Column
    def Metadata(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def ColumnStart(builder: flatbuffers.Builder):
    builder.StartObject(11)

def Start(builder: flatbuffers.Builder):
    ColumnStart(builder)

def ColumnAddName(builder: flatbuffers.Builder, name: int):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)

def AddName(builder: flatbuffers.Builder, name: int):
    ColumnAddName(builder, name)

def ColumnAddType(builder: flatbuffers.Builder, type: int):
    builder.PrependUint8Slot(1, type, 0)

def AddType(builder: flatbuffers.Builder, type: int):
    ColumnAddType(builder, type)

def ColumnAddTitle(builder: flatbuffers.Builder, title: int):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(title), 0)

def AddTitle(builder: flatbuffers.Builder, title: int):
    ColumnAddTitle(builder, title)

def ColumnAddDescription(builder: flatbuffers.Builder, description: int):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(description), 0)

def AddDescription(builder: flatbuffers.Builder, description: int):
    ColumnAddDescription(builder, description)

def ColumnAddWidth(builder: flatbuffers.Builder, width: int):
    builder.PrependInt32Slot(4, width, -1)

def AddWidth(builder: flatbuffers.Builder, width: int):
    ColumnAddWidth(builder, width)

def ColumnAddPrecision(builder: flatbuffers.Builder, precision: int):
    builder.PrependInt32Slot(5, precision, -1)

def AddPrecision(builder: flatbuffers.Builder, precision: int):
    ColumnAddPrecision(builder, precision)

def ColumnAddScale(builder: flatbuffers.Builder, scale: int):
    builder.PrependInt32Slot(6, scale, -1)

def AddScale(builder: flatbuffers.Builder, scale: int):
    ColumnAddScale(builder, scale)

def ColumnAddNullable(builder: flatbuffers.Builder, nullable: bool):
    builder.PrependBoolSlot(7, nullable, 1)

def AddNullable(builder: flatbuffers.Builder, nullable: bool):
    ColumnAddNullable(builder, nullable)

def ColumnAddUnique(builder: flatbuffers.Builder, unique: bool):
    builder.PrependBoolSlot(8, unique, 0)

def AddUnique(builder: flatbuffers.Builder, unique: bool):
    ColumnAddUnique(builder, unique)

def ColumnAddPrimaryKey(builder: flatbuffers.Builder, primaryKey: bool):
    builder.PrependBoolSlot(9, primaryKey, 0)

def AddPrimaryKey(builder: flatbuffers.Builder, primaryKey: bool):
    ColumnAddPrimaryKey(builder, primaryKey)

def ColumnAddMetadata(builder: flatbuffers.Builder, metadata: int):
    builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(metadata), 0)

def AddMetadata(builder: flatbuffers.Builder, metadata: int):
    ColumnAddMetadata(builder, metadata)

def ColumnEnd(builder: flatbuffers.Builder) -> int:
    return builder.EndObject()

def End(builder: flatbuffers.Builder) -> int:
    return ColumnEnd(builder)



from geojson import Feature as GeoJsonFeature

from flatgeobuf.FlatGeobuf.Feature import Feature
from flatgeobuf.generic.feature import BaseFeature, parse_properties
from flatgeobuf.geojson.geometry import from_geometry
from flatgeobuf.header_meta import HeaderMeta


class IGeoJsonFeature(BaseFeature, GeoJsonFeature):
    pass


def from_feature(feature: Feature, header: HeaderMeta) -> IGeoJsonFeature:
    columns = header.columns
    _geometry = feature.Geometry()

    if not _geometry:
        geometry = None
    else:
        geometry = from_geometry(_geometry, header.geometry_type)

    geojson_feature = GeoJsonFeature(
        type="Feature",
        geometry=geometry,
        properties=parse_properties(feature, columns),
    )
    return geojson_feature



from dataclasses import dataclass
from typing import Optional

from flatgeobuf.FlatGeobuf.ColumnType import ColumnType


@dataclass
class ColumnMeta:
    name: str
    type: ColumnType
    title: Optional[str] = None
    description: Optional[str] = None
    width: int = 0
    precision: int = 0
    scale: int = 0
    nullable: bool = False
    unique: bool = False
    primary_key: bool = False



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FlatGeobuf

import flatbuffers
from flatbuffers.compat import import_numpy
from typing import Any
from .Column import Column
from .Crs import Crs
from typing import Optional
np = import_numpy()

class Header(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset: int = 0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Header()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsHeader(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Header
    def Init(self, buf: bytes, pos: int):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Header
    def Name(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Header
    def Envelope(self, j: int):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Header
    def EnvelopeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float64Flags, o)
        return 0

    # Header
    def EnvelopeLength(self) -> int:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Header
    def EnvelopeIsNone(self) -> bool:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Header
    def GeometryType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Header
    def HasZ(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Header
    def HasM(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Header
    def HasT(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Header
    def HasTm(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Header
    def Columns(self, j: int) -> Optional[Column]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            obj = Column()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Header
    def ColumnsLength(self) -> int:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Header
    def ColumnsIsNone(self) -> bool:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # Header
    def FeaturesCount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Header
    def IndexNodeSize(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 16

    # Header
    def Crs(self) -> Optional[Crs]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            obj = Crs()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Header
    def Title(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Header
    def Description(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Header
    def Metadata(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def HeaderStart(builder: flatbuffers.Builder):
    builder.StartObject(14)

def Start(builder: flatbuffers.Builder):
    HeaderStart(builder)

def HeaderAddName(builder: flatbuffers.Builder, name: int):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)

def AddName(builder: flatbuffers.Builder, name: int):
    HeaderAddName(builder, name)

def HeaderAddEnvelope(builder: flatbuffers.Builder, envelope: int):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(envelope), 0)

def AddEnvelope(builder: flatbuffers.Builder, envelope: int):
    HeaderAddEnvelope(builder, envelope)

def HeaderStartEnvelopeVector(builder, numElems: int) -> int:
    return builder.StartVector(8, numElems, 8)

def StartEnvelopeVector(builder, numElems: int) -> int:
    return HeaderStartEnvelopeVector(builder, numElems)

def HeaderAddGeometryType(builder: flatbuffers.Builder, geometryType: int):
    builder.PrependUint8Slot(2, geometryType, 0)

def AddGeometryType(builder: flatbuffers.Builder, geometryType: int):
    HeaderAddGeometryType(builder, geometryType)

def HeaderAddHasZ(builder: flatbuffers.Builder, hasZ: bool):
    builder.PrependBoolSlot(3, hasZ, 0)

def AddHasZ(builder: flatbuffers.Builder, hasZ: bool):
    HeaderAddHasZ(builder, hasZ)

def HeaderAddHasM(builder: flatbuffers.Builder, hasM: bool):
    builder.PrependBoolSlot(4, hasM, 0)

def AddHasM(builder: flatbuffers.Builder, hasM: bool):
    HeaderAddHasM(builder, hasM)

def HeaderAddHasT(builder: flatbuffers.Builder, hasT: bool):
    builder.PrependBoolSlot(5, hasT, 0)

def AddHasT(builder: flatbuffers.Builder, hasT: bool):
    HeaderAddHasT(builder, hasT)

def HeaderAddHasTm(builder: flatbuffers.Builder, hasTm: bool):
    builder.PrependBoolSlot(6, hasTm, 0)

def AddHasTm(builder: flatbuffers.Builder, hasTm: bool):
    HeaderAddHasTm(builder, hasTm)

def HeaderAddColumns(builder: flatbuffers.Builder, columns: int):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(columns), 0)

def AddColumns(builder: flatbuffers.Builder, columns: int):
    HeaderAddColumns(builder, columns)

def HeaderStartColumnsVector(builder, numElems: int) -> int:
    return builder.StartVector(4, numElems, 4)

def StartColumnsVector(builder, numElems: int) -> int:
    return HeaderStartColumnsVector(builder, numElems)

def HeaderAddFeaturesCount(builder: flatbuffers.Builder, featuresCount: int):
    builder.PrependUint64Slot(8, featuresCount, 0)

def AddFeaturesCount(builder: flatbuffers.Builder, featuresCount: int):
    HeaderAddFeaturesCount(builder, featuresCount)

def HeaderAddIndexNodeSize(builder: flatbuffers.Builder, indexNodeSize: int):
    builder.PrependUint16Slot(9, indexNodeSize, 16)

def AddIndexNodeSize(builder: flatbuffers.Builder, indexNodeSize: int):
    HeaderAddIndexNodeSize(builder, indexNodeSize)

def HeaderAddCrs(builder: flatbuffers.Builder, crs: int):
    builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(crs), 0)

def AddCrs(builder: flatbuffers.Builder, crs: int):
    HeaderAddCrs(builder, crs)

def HeaderAddTitle(builder: flatbuffers.Builder, title: int):
    builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(title), 0)

def AddTitle(builder: flatbuffers.Builder, title: int):
    HeaderAddTitle(builder, title)

def HeaderAddDescription(builder: flatbuffers.Builder, description: int):
    builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(description), 0)

def AddDescription(builder: flatbuffers.Builder, description: int):
    HeaderAddDescription(builder, description)

def HeaderAddMetadata(builder: flatbuffers.Builder, metadata: int):
    builder.PrependUOffsetTRelativeSlot(13, flatbuffers.number_types.UOffsetTFlags.py_type(metadata), 0)

def AddMetadata(builder: flatbuffers.Builder, metadata: int):
    HeaderAddMetadata(builder, metadata)

def HeaderEnd(builder: flatbuffers.Builder) -> int:
    return builder.EndObject()

def End(builder: flatbuffers.Builder) -> int:
    return HeaderEnd(builder)



from __future__ import annotations


class Config:
    global_instance: Config

    def __init__(self):
        self._extra_request_threshold = 256 * 1024

    def extra_request_threshold(self):
        return self._extra_request_threshold

    def set_extra_request_threshold(self, bytes):
        if bytes < 0:
            raise ValueError("extra_request_threshold cannot be negative")
        self._extra_request_threshold = bytes


Config.global_instance = Config()



from __future__ import annotations

from asyncio import StreamReader
from io import BufferedIOBase
from logging import getLogger
from typing import Any, AsyncGenerator, Callable, Generator, Optional, Union

from flatgeobuf.async_http_reader import AsyncHTTPReader
from flatgeobuf.file_reader import FileReader
from flatgeobuf.FlatGeobuf.Feature import Feature
from flatgeobuf.generic.feature import BaseFeature
from flatgeobuf.header_meta import HeaderMeta
from flatgeobuf.http_reader import HTTPReader
from flatgeobuf.packedrtree import Rect

logger = getLogger(__name__)

FromFeatureFn = Callable[[Feature, HeaderMeta], BaseFeature]
ReadFn = Callable[[int, str], Union[bytes, bytearray]]
HeaderMetaFn = Callable[[HeaderMeta], None]


# def deserialize(
#     data: BufferedIOBase,
#     from_feature: FromFeatureFn,
#     header_meta_fn: HeaderMetaFn | None = None,
# ) -> List[BaseFeature]:
#     """Deserialize a FlatGeobuf byte stream to a list of BaseFeature."""

#     if not data.read(3) == magicbytes[:3]:
#         raise ValueError("Not a FlatGeobuf file")

#     bb = data
#     bb.seek(len(magicbytes))
#     header_length = int.from_bytes(bb.read(4), "little")
#     bb.seek(len(magicbytes) + SIZE_PREFIX_LEN)

#     header_bytes = bb.read(header_length)
#     header_meta = from_byte_buffer(header_bytes)

#     logger.debug(f"header_meta: {header_meta}")

#     if header_meta_fn:
#         header_meta_fn(header_meta)

#     offset = len(magicbytes) + SIZE_PREFIX_LEN + header_length
#     index_node_size = header_meta.index_node_size
#     features_count = header_meta.features_count

#     if index_node_size > 0:
#         offset += calc_tree_size(features_count, index_node_size)

#     features = []

#     # Get the buffer size
#     bb.seek(0, os.SEEK_END)
#     buffer_size = bb.tell()
#     logger.debug(f"offset: {offset}, buffer_size: {buffer_size}")

#     while offset < buffer_size:
#         bb.seek(offset)
#         feature_length = int.from_bytes(bb.read(4), "little")

#         bb.seek(offset + SIZE_PREFIX_LEN)
#         feature_bytes = bb.read(feature_length)
#         feature = Feature.GetRootAsFeature(feature_bytes)
#         features.append(from_feature(feature, header_meta))

#         offset += SIZE_PREFIX_LEN + feature_length

#     return features


def deserialize(
    data: BufferedIOBase,
    rect: Rect | None,
    from_feature: FromFeatureFn,
    header_meta_fn: HeaderMetaFn | None = None,
) -> Generator[Any, None, None]:
    """Deserialize a FlatGeobuf byte stream to a list of BaseFeature."""

    reader = FileReader.load(data)

    logger.debug("opened reader")

    if header_meta_fn:
        header_meta_fn(reader.header)

    if not rect:
        rect = [-float("inf"), -float("inf"), float("inf"), float("inf")]

    for feature in reader.select_bbox(rect):
        yield from_feature(feature, reader.header)


async def deserialize_stream(
    stream: StreamReader,
    from_feature: FromFeatureFn,
    header_meta_fn: HeaderMetaFn | None = None,
) -> AsyncGenerator[Any, None]:
    """Deserialize a FlatGeobuf async byte stream to a list of BaseFeature."""

    raise NotImplementedError()
    # reader = slice(stream)

    # async def read(size: int, purpose: str) -> bytes:
    #     return await reader.slice(size)

    # bytes_data = bytes(await read(8, "magic bytes"))
    # if not all(bytes_data[:3] == magicbytes[:3]):
    #     raise ValueError("Not a FlatGeobuf file")

    # bytes_data = bytes(await read(4, "header length"))
    # bb = flatbuffers.ByteBuffer(bytes_data)

    # header_length = bb.readUint32(0)
    # bytes_data = bytes(await read(header_length, "header data"))
    # bb = flatbuffers.ByteBuffer(bytes_data)

    # header_meta = from_byte_buffer(bb)
    # if header_meta_fn:
    #     header_meta_fn(header_meta)

    # index_node_size = header_meta.index_node_size
    # features_count = header_meta.features_count
    # if index_node_size > 0:
    #     tree_size = calc_tree_size(features_count, index_node_size)
    #     await read(tree_size, "entire index, w/o rect")

    # feature = None
    # while feature := await read_feature(read, header_meta, from_feature):
    #     yield feature


async def deserialize_http_async(
    url: str,
    rect: Rect | None,
    from_feature: FromFeatureFn,
    header_meta_fn: HeaderMetaFn | None = None,
) -> AsyncGenerator[Any, None]:
    """Deserialize a FlatGeobuf HTTP resource to a list of BaseFeature."""

    reader = await AsyncHTTPReader.open(url)

    logger.debug("opened reader")

    if header_meta_fn:
        header_meta_fn(reader.header)

    if not rect:
        rect = [-float("inf"), -float("inf"), float("inf"), float("inf")]

    async for feature in reader.select_bbox(rect):
        yield from_feature(feature, reader.header)


def deserialize_http(
    url: str,
    rect: Rect | None,
    from_feature: FromFeatureFn,
    header_meta_fn: HeaderMetaFn | None = None,
) -> Generator[Any, None, None]:
    """Deserialize a FlatGeobuf HTTP resource to a list of BaseFeature."""

    reader = HTTPReader.open(url)

    logger.debug("opened reader")

    if header_meta_fn:
        header_meta_fn(reader.header)

    if not rect:
        rect = [-float("inf"), -float("inf"), float("inf"), float("inf")]

    for feature in reader.select_bbox(rect):
        yield from_feature(feature, reader.header)


async def read_feature(
    read: ReadFn, header_meta: HeaderMeta, from_feature: FromFeatureFn
) -> Optional[BaseFeature]:
    raise NotImplementedError()

    # bytes = bytes(await read(4, "feature length"))
    # if len(bytes) == 0:
    #     return None
    # bb = flatbuffers.ByteBuffer(bytes)
    # feature_length = bb.readUint32(0)
    # bytes = bytes(await read(feature_length, "feature data"))
    # bytes_aligned = bytes(feature_length + 4)
    # bytes_aligned[4:] = bytes
    # bb = flatbuffers.ByteBuffer(bytes_aligned)
    # bb.seek(SIZE_PREFIX_LEN)
    # feature = Feature.GetRootAsFeature(bb)

    # return from_feature(feature, header_meta)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FlatGeobuf

import flatbuffers
from flatbuffers.compat import import_numpy
from typing import Any
from .Column import Column
from .Geometry import Geometry
from typing import Optional
np = import_numpy()

class Feature(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset: int = 0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Feature()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsFeature(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Feature
    def Init(self, buf: bytes, pos: int):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Feature
    def Geometry(self) -> Optional[Geometry]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            obj = Geometry()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Feature
    def Properties(self, j: int):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # Feature
    def PropertiesAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Feature
    def PropertiesLength(self) -> int:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Feature
    def PropertiesIsNone(self) -> bool:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Feature
    def Columns(self, j: int) -> Optional[Column]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            obj = Column()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Feature
    def ColumnsLength(self) -> int:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Feature
    def ColumnsIsNone(self) -> bool:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def FeatureStart(builder: flatbuffers.Builder):
    builder.StartObject(3)

def Start(builder: flatbuffers.Builder):
    FeatureStart(builder)

def FeatureAddGeometry(builder: flatbuffers.Builder, geometry: int):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(geometry), 0)

def AddGeometry(builder: flatbuffers.Builder, geometry: int):
    FeatureAddGeometry(builder, geometry)

def FeatureAddProperties(builder: flatbuffers.Builder, properties: int):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(properties), 0)

def AddProperties(builder: flatbuffers.Builder, properties: int):
    FeatureAddProperties(builder, properties)

def FeatureStartPropertiesVector(builder, numElems: int) -> int:
    return builder.StartVector(1, numElems, 1)

def StartPropertiesVector(builder, numElems: int) -> int:
    return FeatureStartPropertiesVector(builder, numElems)

def FeatureAddColumns(builder: flatbuffers.Builder, columns: int):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(columns), 0)

def AddColumns(builder: flatbuffers.Builder, columns: int):
    FeatureAddColumns(builder, columns)

def FeatureStartColumnsVector(builder, numElems: int) -> int:
    return builder.StartVector(4, numElems, 4)

def StartColumnsVector(builder, numElems: int) -> int:
    return FeatureStartColumnsVector(builder, numElems)

def FeatureEnd(builder: flatbuffers.Builder) -> int:
    return builder.EndObject()

def End(builder: flatbuffers.Builder) -> int:
    return FeatureEnd(builder)



from __future__ import annotations

from io import BufferedIOBase

from geojson import FeatureCollection

from flatgeobuf.generic import HeaderMetaFn
from flatgeobuf.geojson.featurecollection import (
    deserialize,
    deserialize_http,
    deserialize_http_async,
)
from flatgeobuf.packedrtree import Rect


def load(file: BufferedIOBase, *, bbox: Rect | None = None) -> FeatureCollection:
    reader = Reader(file, bbox=bbox)
    features = list(reader)
    return FeatureCollection(features)


async def load_http_async(url: str, *, bbox: Rect | None = None) -> FeatureCollection:
    reader = HTTPReader(url, bbox=bbox)
    features = [feature async for feature in reader]
    return FeatureCollection(features)


def load_http(url: str, *, bbox: Rect | None = None) -> FeatureCollection:
    reader = HTTPReader(url, bbox=bbox)
    features = [feature for feature in reader]
    return FeatureCollection(features)


class Reader:
    def __init__(
        self,
        file: BufferedIOBase,
        *,
        bbox: Rect | None = None,
        header_meta_fn: HeaderMetaFn | None = None,
    ):
        self.file = file
        self.rect = bbox
        self.header_meta_fn = header_meta_fn

    def __iter__(self):
        for feature in deserialize(self.file, self.rect):
            yield feature


class HTTPReader:
    def __init__(
        self,
        url: str,
        *,
        bbox: Rect | None = None,
        header_meta_fn: HeaderMetaFn | None = None,
    ):
        self.url = url
        self.rect = bbox
        self.header_meta_fn = header_meta_fn

    async def __aiter__(self):
        async for feature in deserialize_http_async(self.url, self.rect):
            yield feature

    def __iter__(self):
        for feature in deserialize_http(self.url, self.rect):
            yield feature



from __future__ import annotations

from io import BufferedIOBase
from typing import AsyncGenerator, Union

from geojson import FeatureCollection

from flatgeobuf.generic import HeaderMetaFn
from flatgeobuf.geojson.feature import IGeoJsonFeature
from flatgeobuf.geojson.featurecollection import deserialize as fc_deserialize
from flatgeobuf.geojson.featurecollection import deserialize_http as fc_deserialize_http
from flatgeobuf.geojson.featurecollection import (
    deserialize_stream as fc_deserialize_stream,
)
from flatgeobuf.geojson.reader import HTTPReader  # noqa: F401
from flatgeobuf.geojson.reader import Reader  # noqa: F401
from flatgeobuf.geojson.reader import load  # noqa: F401
from flatgeobuf.geojson.reader import load_http  # noqa: F401
from flatgeobuf.packedrtree import Rect


def deserialize(
    input: Union[BufferedIOBase, str, AsyncGenerator[IGeoJsonFeature, None]],
    rect: Rect | None = None,
    header_meta_fn: HeaderMetaFn | None = None,
) -> Union[FeatureCollection, AsyncGenerator[IGeoJsonFeature, None]]:
    if isinstance(input, BufferedIOBase):
        return fc_deserialize(input, header_meta_fn)
    elif isinstance(input, str):
        return fc_deserialize_http(input, rect, header_meta_fn)
    else:
        return fc_deserialize_stream(input, header_meta_fn)



from __future__ import annotations

from asyncio import StreamReader
from io import BufferedIOBase
from typing import AsyncGenerator, Generator

from geojson import Feature

from flatgeobuf.bbox_filter import BBoxFilter
from flatgeobuf.generic import HeaderMetaFn
from flatgeobuf.generic.featurecollection import deserialize as generic_deserialize
from flatgeobuf.generic.featurecollection import (
    deserialize_http as generic_deserialize_http,
)
from flatgeobuf.generic.featurecollection import (
    deserialize_http_async as generic_deserialize_http_async,
)

# from flatgeobuf.generic.featurecollection import (
#     deserialize_stream as generic_deserialize_stream,
# )
from flatgeobuf.geojson.feature import from_feature
from flatgeobuf.packedrtree import Rect


def deserialize(
    data: BufferedIOBase,
    rect: Rect | None = None,
    header_meta_fn: HeaderMetaFn | None = None,
) -> Generator[Feature, None, None]:
    """Deserialize a FlatGeobuf byte stream to a GeoJSON FeatureCollection."""

    if rect:
        bbox_filter = BBoxFilter(rect)
        for feature in generic_deserialize(data, rect, from_feature, header_meta_fn):
            if bbox_filter.has_intersection(feature):
                yield feature
    else:
        for feature in generic_deserialize(data, rect, from_feature, header_meta_fn):
            yield feature


async def deserialize_stream(
    stream: StreamReader, header_meta_fn: HeaderMetaFn | None = None
) -> AsyncGenerator[Feature, None]:
    """Deserialize a FlatGeobuf async byte stream to a GeoJSON FeatureCollection."""

    raise NotImplementedError()
    # return generic_deserialize_stream(stream, from_feature, header_meta_fn)


async def deserialize_http_async(
    url: str, rect: Rect | None = None, header_meta_fn: HeaderMetaFn | None = None
) -> AsyncGenerator[Feature, None]:
    """Deserialize a FlatGeobuf HTTP resource to a GeoJSON FeatureCollection."""

    if rect:
        bbox_filter = BBoxFilter(rect)
        async for feature in generic_deserialize_http_async(
            url, rect, from_feature, header_meta_fn
        ):
            if bbox_filter.has_intersection(feature):
                yield feature
    else:
        async for feature in generic_deserialize_http_async(
            url, rect, from_feature, header_meta_fn
        ):
            yield feature


def deserialize_http(
    url: str, rect: Rect | None = None, header_meta_fn: HeaderMetaFn | None = None
) -> Generator[Feature, None, None]:
    """Deserialize a FlatGeobuf HTTP resource to a GeoJSON FeatureCollection."""

    if rect:
        bbox_filter = BBoxFilter(rect)
        for feature in generic_deserialize_http(
            url, rect, from_feature, header_meta_fn
        ):
            if bbox_filter.has_intersection(feature):
                yield feature
    else:
        for feature in generic_deserialize_http(
            url, rect, from_feature, header_meta_fn
        ):
            yield feature



from __future__ import annotations

from dataclasses import dataclass
from typing import List

from flatgeobuf.column_meta import ColumnMeta
from flatgeobuf.crs_meta import CrsMeta
from flatgeobuf.FlatGeobuf.GeometryType import GeometryType
from flatgeobuf.FlatGeobuf.Header import Header


@dataclass
class HeaderMeta:
    geometry_type: GeometryType
    columns: List[ColumnMeta] | None
    envelope: List[float] | None  # Float64Array
    features_count: int
    index_node_size: int
    crs: CrsMeta | None
    title: str | None
    description: str | None
    metadata: str | None


def from_byte_buffer(bb: bytes | bytearray) -> HeaderMeta:
    header = Header.GetRootAsHeader(bb, 0)
    features_count = header.FeaturesCount()
    index_node_size = header.IndexNodeSize()

    columns = []
    for j in range(header.ColumnsLength()):
        column = header.Columns(j)
        if not column:
            raise ValueError("Column unexpectedly missing")
        # NOTE: FlatBuf returns bytes, not str
        column_name = column.Name()
        if not column_name:
            raise ValueError("Column name unexpectedly missing")
        columns.append(
            ColumnMeta(
                name=column_name.decode("utf-8"),
                type=column.Type(),
                title=column.Title(),
                description=column.Description(),
                width=column.Width(),
                precision=column.Precision(),
                scale=column.Scale(),
                nullable=column.Nullable(),
                unique=column.Unique(),
                primary_key=column.PrimaryKey(),
            )
        )

    crs = header.Crs()
    crs_meta = None
    if crs:
        crs_meta = CrsMeta(
            org=crs.Org(),
            code=crs.Code(),
            name=crs.Name(),
            description=crs.Description(),
            wkt=crs.Wkt(),
            code_string=crs.CodeString(),
        )

    header_meta = HeaderMeta(
        geometry_type=header.GeometryType(),
        columns=columns,
        envelope=None,
        features_count=features_count,
        index_node_size=index_node_size,
        crs=crs_meta,
        title=header.Title(),
        description=header.Description(),
        metadata=header.Metadata(),
    )

    return header_meta



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FlatGeobuf

import flatbuffers
from flatbuffers.compat import import_numpy
from typing import Any
from typing import Optional
np = import_numpy()

class Geometry(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset: int = 0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Geometry()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsGeometry(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Geometry
    def Init(self, buf: bytes, pos: int):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Geometry
    def Ends(self, j: int):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Geometry
    def EndsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint32Flags, o)
        return 0

    # Geometry
    def EndsLength(self) -> int:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Geometry
    def EndsIsNone(self) -> bool:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Geometry
    def Xy(self, j: int):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Geometry
    def XyAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float64Flags, o)
        return 0

    # Geometry
    def XyLength(self) -> int:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Geometry
    def XyIsNone(self) -> bool:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Geometry
    def Z(self, j: int):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Geometry
    def ZAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float64Flags, o)
        return 0

    # Geometry
    def ZLength(self) -> int:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Geometry
    def ZIsNone(self) -> bool:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Geometry
    def M(self, j: int):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Geometry
    def MAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float64Flags, o)
        return 0

    # Geometry
    def MLength(self) -> int:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Geometry
    def MIsNone(self) -> bool:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Geometry
    def T(self, j: int):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Geometry
    def TAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float64Flags, o)
        return 0

    # Geometry
    def TLength(self) -> int:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Geometry
    def TIsNone(self) -> bool:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Geometry
    def Tm(self, j: int):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Geometry
    def TmAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # Geometry
    def TmLength(self) -> int:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Geometry
    def TmIsNone(self) -> bool:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Geometry
    def Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Geometry
    def Parts(self, j: int) -> Optional['Geometry']:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            obj = Geometry()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Geometry
    def PartsLength(self) -> int:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Geometry
    def PartsIsNone(self) -> bool:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

def GeometryStart(builder: flatbuffers.Builder):
    builder.StartObject(8)

def Start(builder: flatbuffers.Builder):
    GeometryStart(builder)

def GeometryAddEnds(builder: flatbuffers.Builder, ends: int):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(ends), 0)

def AddEnds(builder: flatbuffers.Builder, ends: int):
    GeometryAddEnds(builder, ends)

def GeometryStartEndsVector(builder, numElems: int) -> int:
    return builder.StartVector(4, numElems, 4)

def StartEndsVector(builder, numElems: int) -> int:
    return GeometryStartEndsVector(builder, numElems)

def GeometryAddXy(builder: flatbuffers.Builder, xy: int):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(xy), 0)

def AddXy(builder: flatbuffers.Builder, xy: int):
    GeometryAddXy(builder, xy)

def GeometryStartXyVector(builder, numElems: int) -> int:
    return builder.StartVector(8, numElems, 8)

def StartXyVector(builder, numElems: int) -> int:
    return GeometryStartXyVector(builder, numElems)

def GeometryAddZ(builder: flatbuffers.Builder, z: int):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(z), 0)

def AddZ(builder: flatbuffers.Builder, z: int):
    GeometryAddZ(builder, z)

def GeometryStartZVector(builder, numElems: int) -> int:
    return builder.StartVector(8, numElems, 8)

def StartZVector(builder, numElems: int) -> int:
    return GeometryStartZVector(builder, numElems)

def GeometryAddM(builder: flatbuffers.Builder, m: int):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(m), 0)

def AddM(builder: flatbuffers.Builder, m: int):
    GeometryAddM(builder, m)

def GeometryStartMVector(builder, numElems: int) -> int:
    return builder.StartVector(8, numElems, 8)

def StartMVector(builder, numElems: int) -> int:
    return GeometryStartMVector(builder, numElems)

def GeometryAddT(builder: flatbuffers.Builder, t: int):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(t), 0)

def AddT(builder: flatbuffers.Builder, t: int):
    GeometryAddT(builder, t)

def GeometryStartTVector(builder, numElems: int) -> int:
    return builder.StartVector(8, numElems, 8)

def StartTVector(builder, numElems: int) -> int:
    return GeometryStartTVector(builder, numElems)

def GeometryAddTm(builder: flatbuffers.Builder, tm: int):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(tm), 0)

def AddTm(builder: flatbuffers.Builder, tm: int):
    GeometryAddTm(builder, tm)

def GeometryStartTmVector(builder, numElems: int) -> int:
    return builder.StartVector(8, numElems, 8)

def StartTmVector(builder, numElems: int) -> int:
    return GeometryStartTmVector(builder, numElems)

def GeometryAddType(builder: flatbuffers.Builder, type: int):
    builder.PrependUint8Slot(6, type, 0)

def AddType(builder: flatbuffers.Builder, type: int):
    GeometryAddType(builder, type)

def GeometryAddParts(builder: flatbuffers.Builder, parts: int):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(parts), 0)

def AddParts(builder: flatbuffers.Builder, parts: int):
    GeometryAddParts(builder, parts)

def GeometryStartPartsVector(builder, numElems: int) -> int:
    return builder.StartVector(4, numElems, 4)

def StartPartsVector(builder, numElems: int) -> int:
    return GeometryStartPartsVector(builder, numElems)

def GeometryEnd(builder: flatbuffers.Builder) -> int:
    return builder.EndObject()

def End(builder: flatbuffers.Builder) -> int:
    return GeometryEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FlatGeobuf

import flatbuffers
from flatbuffers.compat import import_numpy
from typing import Any
from typing import Optional
np = import_numpy()

class Crs(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset: int = 0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Crs()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsCrs(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Crs
    def Init(self, buf: bytes, pos: int):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Crs
    def Org(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Crs
    def Code(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Crs
    def Name(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Crs
    def Description(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Crs
    def Wkt(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Crs
    def CodeString(self) -> Optional[str]:
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def CrsStart(builder: flatbuffers.Builder):
    builder.StartObject(6)

def Start(builder: flatbuffers.Builder):
    CrsStart(builder)

def CrsAddOrg(builder: flatbuffers.Builder, org: int):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(org), 0)

def AddOrg(builder: flatbuffers.Builder, org: int):
    CrsAddOrg(builder, org)

def CrsAddCode(builder: flatbuffers.Builder, code: int):
    builder.PrependInt32Slot(1, code, 0)

def AddCode(builder: flatbuffers.Builder, code: int):
    CrsAddCode(builder, code)

def CrsAddName(builder: flatbuffers.Builder, name: int):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)

def AddName(builder: flatbuffers.Builder, name: int):
    CrsAddName(builder, name)

def CrsAddDescription(builder: flatbuffers.Builder, description: int):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(description), 0)

def AddDescription(builder: flatbuffers.Builder, description: int):
    CrsAddDescription(builder, description)

def CrsAddWkt(builder: flatbuffers.Builder, wkt: int):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(wkt), 0)

def AddWkt(builder: flatbuffers.Builder, wkt: int):
    CrsAddWkt(builder, wkt)

def CrsAddCodeString(builder: flatbuffers.Builder, codeString: int):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(codeString), 0)

def AddCodeString(builder: flatbuffers.Builder, codeString: int):
    CrsAddCodeString(builder, codeString)

def CrsEnd(builder: flatbuffers.Builder) -> int:
    return builder.EndObject()

def End(builder: flatbuffers.Builder) -> int:
    return CrsEnd(builder)



from dataclasses import dataclass
from typing import Optional

@dataclass
class CrsMeta:
    org: Optional[str] = None
    code: int = 0
    name: Optional[str] = None
    description: Optional[str] = None
    wkt: Optional[str] = None
    code_string: Optional[str] = None


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FlatGeobuf


class GeometryType(object):
    Unknown = 0
    Point = 1
    LineString = 2
    Polygon = 3
    MultiPoint = 4
    MultiLineString = 5
    MultiPolygon = 6
    GeometryCollection = 7
    CircularString = 8
    CompoundCurve = 9
    CurvePolygon = 10
    MultiCurve = 11
    MultiSurface = 12
    Curve = 13
    Surface = 14
    PolyhedralSurface = 15
    TIN = 16
    Triangle = 17



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FlatGeobuf

class ColumnType(object):
    Byte = 0
    UByte = 1
    Bool = 2
    Short = 3
    UShort = 4
    Int = 5
    UInt = 6
    Long = 7
    ULong = 8
    Float = 9
    Double = 10
    String = 11
    Json = 12
    DateTime = 13
    Binary = 14



from abc import ABCMeta
import json
from typing import Any, Callable, Dict, List, Optional, Union
from flatgeobuf.FlatGeobuf.GeometryType import GeometryType

from flatgeobuf.column_meta import ColumnMeta
from flatgeobuf.FlatGeobuf.ColumnType import ColumnType
from flatgeobuf.FlatGeobuf.Feature import Feature
from flatgeobuf.header_meta import HeaderMeta

from flatgeobuf.generic.geometry import ParsedGeometry, SimpleGeometry


class BaseFeature(metaclass=ABCMeta):
    def get_geometry(self) -> Optional[SimpleGeometry]:
        pass

    def get_properties(self) -> Any:
        pass

    def set_properties(self, properties: Dict[str, Union[bool, int, str, Any]]) -> Any:
        pass


CreateFeatureCallable = Callable[
    [SimpleGeometry, Dict[str, Union[bool, int, str, Any]]], BaseFeature
]

IProperties = Dict[str, Union[bool, int, str, Any]]


def from_feature(
    feature: Feature,
    header: HeaderMeta,
    create_geometry: Callable[[ParsedGeometry, GeometryType], SimpleGeometry],
    create_feature: Callable[
        [SimpleGeometry, Dict[str, Union[bool, int, str, Any]]], BaseFeature
    ],
) -> BaseFeature:
    columns = header.columns
    geometry = feature.Geometry()
    if not geometry:
        raise ValueError("Feature has no geometry")
    simple_geometry = create_geometry(geometry, header.geometry_type)
    properties = parse_properties(feature, columns)
    return create_feature(simple_geometry, properties)


def parse_properties(
    feature: Feature, columns: Optional[List[ColumnMeta]] = None
) -> Dict[str, Union[bool, int, str, Any]]:
    properties = {}
    if not columns or len(columns) == 0:
        return properties
    array = feature.PropertiesAsNumpy().tobytes()
    if array == 0:
        return properties
    view = memoryview(array)
    length = feature.PropertiesLength()
    offset = 0
    while offset < length:
        i = view[offset : offset + 2].cast("H")[0]
        offset += 2
        column = columns[i]
        name = column.name
        column_type = column.type
        if column_type == ColumnType.Bool:
            properties[name] = view[offset].cast("b")[0] == 1
            offset += 1
        elif column_type == ColumnType.Byte:
            properties[name] = view[offset].cast("b")[0]
            offset += 1
        elif column_type == ColumnType.UByte:
            properties[name] = view[offset].cast("B")[0]
            offset += 1
        elif column_type == ColumnType.Short:
            properties[name] = view[offset : offset + 2].cast("h")[0]
            offset += 2
        elif column_type == ColumnType.UShort:
            properties[name] = view[offset : offset + 2].cast("H")[0]
            offset += 2
        elif column_type == ColumnType.Int:
            properties[name] = view[offset : offset + 4].cast("i")[0]
            offset += 4
        elif column_type == ColumnType.UInt:
            properties[name] = view[offset : offset + 4].cast("I")[0]
            offset += 4
        elif column_type == ColumnType.Long:
            properties[name] = view[offset : offset + 8].cast("q")[0]

            offset += 8
        elif column_type == ColumnType.ULong:
            properties[name] = view[offset : offset + 8].cast("Q")[0]
            offset += 8
        elif column_type == ColumnType.Float:
            properties[name] = view[offset : offset + 4].cast("f")[0]
            offset += 4
        elif column_type == ColumnType.Double:
            properties[name] = view[offset : offset + 8].cast("d")[0]
            offset += 8
        elif column_type in (ColumnType.DateTime, ColumnType.String):
            str_length = view[offset : offset + 4].cast("I")[0]
            offset += 4
            str_value = view[offset : offset + str_length].tobytes().decode()
            properties[name] = str_value
            offset += str_length
        elif column_type == ColumnType.Json:
            str_length = view[offset : offset + 4].cast("I")[0]
            offset += 4
            str_value = view[offset : offset + str_length].tobytes().decode()
            properties[name] = json.loads(str_value)
            offset += str_length
        else:
            raise ValueError(f"Unknown type {column_type}")

    return properties



from __future__ import annotations

from asyncio import StreamReader
from io import BufferedIOBase
from typing import AsyncGenerator, List, Union

from flatgeobuf.FlatGeobuf.ColumnType import ColumnType
from flatgeobuf.FlatGeobuf.GeometryType import GeometryType
from flatgeobuf.generic.feature import BaseFeature
from flatgeobuf.generic.featurecollection import HeaderMetaFn  # noqa: F401
from flatgeobuf.generic.featurecollection import FromFeatureFn
from flatgeobuf.generic.featurecollection import deserialize as deserialize_buffer
from flatgeobuf.generic.featurecollection import deserialize_http, deserialize_stream
from flatgeobuf.packedrtree import Rect


def deserialize(
    input: Union[BufferedIOBase, StreamReader, str],
    from_feature: FromFeatureFn,
    rect: Rect | None = None,
) -> Union[List[BaseFeature], AsyncGenerator[BaseFeature, None]]:
    if isinstance(input, BufferedIOBase):
        return deserialize_buffer(input, from_feature)
    elif isinstance(input, StreamReader):
        return deserialize_stream(input, from_feature)
    else:
        return deserialize_http(input, rect, from_feature)


__all__ = [
    "GeometryType",
    "ColumnType",
    "deserialize",
    # "serialize"
]



from __future__ import annotations

import math
from logging import getLogger
from typing import (
    Annotated,
    AsyncGenerator,
    Awaitable,
    Callable,
    Generator,
    List,
    Tuple,
    Union,
)

from .config import Config

logger = getLogger(__name__)

NODE_ITEM_BYTE_LEN: int = 8 * 4 + 8

DEFAULT_NODE_SIZE = 16

Rect = Union[Tuple[float, float, float, float], Annotated[list[float], 4]]
SearchResult = Tuple[int, int, Union[int, None]]


def calc_tree_size(num_items: int, node_size: int) -> int:
    node_size = min(max(int(node_size), 2), 65535)
    n = num_items
    num_nodes = n
    while n != 1:
        n = math.ceil(n / node_size)
        num_nodes += n
    return num_nodes * NODE_ITEM_BYTE_LEN


def generate_level_bounds(num_items: int, node_size: int) -> List[Tuple[int, int]]:
    if node_size < 2:
        raise ValueError("Node size must be at least 2")
    if num_items == 0:
        raise ValueError("Number of items must be greater than 0")

    n = num_items
    num_nodes = n
    level_num_nodes = [n]
    while n != 1:
        n = math.ceil(n / node_size)
        num_nodes += n
        level_num_nodes.append(n)

    level_offsets = []
    n = num_nodes
    for size in level_num_nodes:
        level_offsets.append(n - size)
        n -= size

    level_bounds = []
    for i in range(len(level_num_nodes)):
        level_bounds.append((level_offsets[i], level_offsets[i] + level_num_nodes[i]))
    return level_bounds


class NodeRange:
    def __init__(self, nodes: Tuple[int, int], level: int):
        self._level = level
        self.nodes = nodes

    def level(self) -> int:
        return self._level

    def start_node_idx(self) -> int:
        return self.nodes[0]

    def end_node_idx(self) -> int:
        return self.nodes[1]

    def extend_end_node_idx(self, new_idx: int):
        assert new_idx > self.nodes[1]
        self.nodes = (self.nodes[0], new_idx)

    def __str__(self) -> str:
        return (
            f"[NodeRange level: {self._level}, nodes: {self.nodes[0]}-{self.nodes[1]}]"
        )


class PackedRTree:
    def __init__(self, num_items: int, node_size: int, rect: Rect):
        self.num_items = num_items
        self.node_size = node_size
        self.rect = rect

        self.min_x, self.min_y, self.max_x, self.max_y = rect
        self.level_bounds = generate_level_bounds(num_items, node_size)
        self.first_leaf_node_idx = self.level_bounds[0][0]
        self.root_node_range = NodeRange((0, 1), len(self.level_bounds) - 1)

        self.queue = [self.root_node_range]

    def _prefetch(self):
        self.node_range = self.queue.pop(0)

        logger.debug(f"popped node: {self.node_range}, queue_length: {len(self.queue)}")

        self.node_range_start_idx = self.node_range.start_node_idx()
        self.is_leaf_node = self.node_range_start_idx >= self.first_leaf_node_idx

        # find the end index of the node
        _node_range_end_idx = self.node_range.end_node_idx()
        level_bound = self.level_bounds[self.node_range.level()][1]
        node_idx = min(_node_range_end_idx + self.node_size, level_bound)
        if self.is_leaf_node and node_idx < level_bound:
            # We can infer the length of *this* feature by getting the start of the *next*
            # feature, so we get an extra node.
            # This approach doesn't work for the final node in the index,
            # but in that case we know that the feature runs to the end of the FGB file and
            # could make an open ended range request to get "the rest of the data".
            self.node_range_end_idx = node_idx + 1
        else:
            self.node_range_end_idx = node_idx

        self.num_nodes_in_range = self.node_range_end_idx - self.node_range_start_idx

    def iterate(self, node_idx: int) -> SearchResult | None:
        node_idx_in_data_view = node_idx - self.node_range_start_idx
        data_view_byte_start = node_idx_in_data_view * NODE_ITEM_BYTE_LEN

        # TODO: Enforce little endian
        if (
            self.max_x
            < self.data_view[data_view_byte_start : data_view_byte_start + 8].cast("d")[
                0
            ]
        ):
            return
        if (
            self.max_y
            < self.data_view[data_view_byte_start + 8 : data_view_byte_start + 16].cast(
                "d"
            )[0]
        ):
            return
        if (
            self.min_x
            > self.data_view[
                data_view_byte_start + 16 : data_view_byte_start + 24
            ].cast("d")[0]
        ):
            return
        if (
            self.min_y
            > self.data_view[
                data_view_byte_start + 24 : data_view_byte_start + 32
            ].cast("d")[0]
        ):
            return

        offset = self.data_view[
            data_view_byte_start + 32 : data_view_byte_start + 40
        ].cast("q")[0]

        if self.is_leaf_node:
            feature_byte_offset = offset
            feature_length = None
            if node_idx < self.num_items - 1:
                next_pos = (node_idx_in_data_view + 1) * NODE_ITEM_BYTE_LEN
                next_offset = self.data_view[next_pos + 32 : next_pos + 40].cast("q")[0]
                feature_length = next_offset - feature_byte_offset
            feature_idx = node_idx - self.first_leaf_node_idx
            return (feature_byte_offset, feature_idx, feature_length)

        first_child_node_idx = offset

        extra_request_threshold_nodes = (
            Config.global_instance.extra_request_threshold() // NODE_ITEM_BYTE_LEN
        )

        nearest_node_range = self.queue[-1] if self.queue else None
        if (
            nearest_node_range
            and nearest_node_range.level() == self.node_range.level() - 1
            and first_child_node_idx
            < nearest_node_range.end_node_idx() + extra_request_threshold_nodes
        ):
            nearest_node_range.extend_end_node_idx(first_child_node_idx)
            return

        new_node_range = NodeRange(
            (first_child_node_idx, first_child_node_idx + 1),
            self.node_range.level() - 1,
        )

        if nearest_node_range and nearest_node_range.level() == new_node_range.level():
            logger.info(
                f"Same level, but too far away. Pushing new request for node_idx: {first_child_node_idx} rather than merging with distant {nearest_node_range}"
            )
        else:
            logger.info(
                f"Pushing new level for {new_node_range} onto queue with nearest_node_range: {nearest_node_range} since there's not already a range for this level."
            )

        self.queue.append(new_node_range)

    async def stream_search_async(
        self, read_node: Callable[[int, int], Awaitable[bytes]]
    ) -> AsyncGenerator[SearchResult, None]:
        while self.queue:
            self._prefetch()

            buffer = await read_node(
                self.node_range_start_idx * NODE_ITEM_BYTE_LEN,
                self.num_nodes_in_range * NODE_ITEM_BYTE_LEN,
            )

            self.data_view = memoryview(buffer)

            for node_idx in range(self.node_range_start_idx, self.node_range_end_idx):
                iteration = self.iterate(node_idx)
                if iteration:
                    yield iteration

    def stream_search(
        self, read_node: Callable[[int, int], bytes]
    ) -> Generator[SearchResult, None, None]:
        while self.queue:
            self._prefetch()

            buffer = read_node(
                self.node_range_start_idx * NODE_ITEM_BYTE_LEN,
                self.num_nodes_in_range * NODE_ITEM_BYTE_LEN,
            )

            self.data_view = memoryview(buffer)

            for node_idx in range(self.node_range_start_idx, self.node_range_end_idx):
                iteration = self.iterate(node_idx)
                if iteration:
                    yield iteration



from __future__ import annotations

from logging import getLogger
from typing import Any, AsyncGenerator, List, Tuple, Generator

from flatgeobuf.config import Config
from flatgeobuf.constants import SIZE_PREFIX_LEN, magicbytes
from flatgeobuf.FlatGeobuf.Feature import Feature
from flatgeobuf.header_meta import HeaderMeta, from_byte_buffer
from flatgeobuf.http_range_client import BufferedHttpRangeClient
from flatgeobuf.packedrtree import (
    DEFAULT_NODE_SIZE,
    NODE_ITEM_BYTE_LEN,
    PackedRTree,
    Rect,
    calc_tree_size,
)

logger = getLogger(__name__)


def merge_promises(
    promises: List[Generator[Feature, Any, None]]
) -> Generator[Feature, None, None]:
    for promise in promises:
        for feature in promise:
            yield feature


class HTTPReader:
    def __init__(
        self,
        header_client: BufferedHttpRangeClient,
        header: HeaderMeta,
        header_length: int,
        index_length: int,
    ):
        self.header_client = header_client
        self.header = header
        self.header_length = header_length
        self.index_length = index_length

    # Fetch the header, preparing the reader to read Feature data.
    # and potentially some opportunistic fetching of the index.
    @staticmethod
    def open(url: str) -> HTTPReader:
        # In reality, the header is probably less than half this size, but
        # better to overshoot and fetch an extra kb rather than have to issue
        # a second request.
        assumed_header_length = 2024

        header_client = BufferedHttpRangeClient(url)

        # Immediately following the header is the optional spatial index, we deliberately fetch
        # a small part of that to skip subsequent requests.
        def assumed_index_length() -> int:
            # The actual branching factor will be in the header, but since we
            # don't have the header yet, we just guess. The consequence of
            # getting this wrong isn't terminal, it only means we may be
            # fetching slightly more than we need or that we need to make an
            # extra request later.
            assumed_branching_factor = DEFAULT_NODE_SIZE

            # NOTE: each layer is exponentially larger
            prefetched_layers = 3

            result = 0
            for i in range(prefetched_layers):
                layer_width = assumed_branching_factor**i * NODE_ITEM_BYTE_LEN
                result += layer_width
            return result

        min_req_length = assumed_header_length + assumed_index_length()
        logger.debug(
            f"fetching header. min_req_length: {min_req_length} (assumed_header_length: {assumed_header_length}, assumed_index_length: {assumed_index_length()})",
        )

        bytes = header_client.get_range(0, 8, min_req_length, "header")
        if not bytes[:3] == magicbytes[:3]:
            logger.error(f"bytes: {bytes} != {magicbytes}")
            raise ValueError("Not a FlatGeobuf file")
        logger.debug("magic bytes look good")

        bytes = header_client.get_range(8, 4, min_req_length, "header")
        header_length = int.from_bytes(bytes, "little")
        HEADER_MAX_BUFFER_SIZE = 1048576 * 10
        if header_length > HEADER_MAX_BUFFER_SIZE or header_length < 8:
            # minimum size check avoids panic in FlatBuffers header decoding
            raise ValueError("Invalid header size")
        logger.debug(f"header_length: {header_length}")

        bytes = header_client.get_range(
            12, header_length, min_req_length, "header"
        )
        bb = bytearray(bytes)  # bb = flatbuffers.ByteBuffer(bytes)
        header = from_byte_buffer(bb)

        index_length = calc_tree_size(header.features_count, header.index_node_size)

        logger.debug("completed: opening http reader")

        return HTTPReader(header_client, header, header_length, index_length)

    def select_bbox(self, rect: Rect) -> Generator[Feature, None, None]:
        # Read R-Tree index and build filter for features within bbox
        length_before_tree = self.length_before_tree()

        buffered_client = self.header_client

        def read_node(offset_into_tree: int, size: int) -> bytes:
            min_req_length = 0
            return buffered_client.get_range(
                length_before_tree + offset_into_tree,
                size,
                min_req_length,
                "index",
            )

        batches: List[List[Tuple[int, int]]] = []
        current_batch: List[Tuple[int, int]] = []

        for search_result in PackedRTree(
            self.header.features_count,
            self.header.index_node_size,
            rect,
        ).stream_search(read_node):
            feature_offset, _, feature_length = search_result
            if not feature_length:
                logger.info("final feature")
                # Normally we get the feature length by subtracting between
                # adjacent nodes from the index, which we can't do for the
                # _very_ last feature in a dataset.
                #
                # We could *guess* the size, but we'd risk overshooting the length,
                # which will cause some webservers to return HTTP 416: Unsatisfiable range
                #
                # So instead we fetch only the final features byte length, stored in the
                # first 4 bytes.
                feature_length = 4

            if not current_batch:
                current_batch.append((feature_offset, feature_length))
                continue

            prev_feature = current_batch[-1]
            gap = feature_offset - (prev_feature[0] + prev_feature[1])

            if gap > Config.global_instance.extra_request_threshold():
                logger.info(
                    f"Pushing new feature batch, since gap {gap} was too large",
                )
                batches.append(current_batch)
                current_batch = []

            current_batch.append((feature_offset, feature_length))

        self.header_client.log_usage("header+index")

        if current_batch:
            batches.append(current_batch)

        promises: List[Generator[Feature, Any, None]] = [
            self.read_feature_batch(batch) for batch in batches
        ]

        # Fetch all batches concurrently, yielding features as they become
        # available, meaning the results may be intermixed.
        for promise in merge_promises(promises):
            yield promise

    def length_before_tree(self) -> int:
        # FGB Layout is: [magicbytes (fixed), headerLength (i32), header (variable), Tree (variable), Features (variable)]
        return len(magicbytes) + SIZE_PREFIX_LEN + self.header_length

    def length_before_features(self) -> int:
        return self.length_before_tree() + self.index_length

    def build_feature_client(self) -> BufferedHttpRangeClient:
        return BufferedHttpRangeClient(self.header_client.http_client)

    def read_feature_batch(
        self, batch: List[Tuple[int, int]]
    ) -> Generator[Feature, None, None]:
        first_feature_offset = batch[0][0]
        last_feature_offset, last_feature_length = batch[-1]

        batch_start = first_feature_offset
        batch_end = last_feature_offset + last_feature_length
        batch_size = batch_end - batch_start

        # A new feature client is needed for each batch to own the underlying buffer as features are yielded.
        feature_client = self.build_feature_client()

        min_feature_req_length = batch_size
        for feature_offset, _ in batch:
            yield self.read_feature(
                feature_client,
                feature_offset,
                min_feature_req_length,
            )
            # Only set min_feature_req_length for the first request.
            #
            # This should only affect a batch that contains the final feature, otherwise
            # we've calculated `batch_size` to get all the data we need for the batch.
            # For the very final feature in a dataset, we don't know it's length, so we
            # will end up executing an extra request for that batch.
            min_feature_req_length = 0
        feature_client.log_usage("feature")

    def read_feature(
        self,
        feature_client: BufferedHttpRangeClient,
        feature_offset: int,
        min_feature_req_length: int,
    ) -> Feature:
        offset = feature_offset + self.length_before_features()

        bytes = feature_client.get_range(
            offset,
            4,
            min_feature_req_length,
            "feature length",
        )
        feature_length = int.from_bytes(bytes, "little")

        byte_buffer = feature_client.get_range(
            offset + 4,
            feature_length,
            min_feature_req_length,
            "feature data",
        )
        bytes = bytearray(byte_buffer)
        bytes_aligned = bytearray(feature_length + SIZE_PREFIX_LEN)
        bytes_aligned[SIZE_PREFIX_LEN:] = bytes
        bb = bytearray(bytes_aligned)
        bb = bb[SIZE_PREFIX_LEN:]
        return Feature.GetRootAsFeature(bb)



from __future__ import annotations

import urllib.request
from logging import getLogger

logger = getLogger(__name__)


class BufferedHttpRangeClient:
    def __init__(self, source: str | HttpRangeClient):
        self.bytes_ever_used = 0
        self.bytes_ever_fetched = 0
        self.buffer = bytearray()
        self.head = 0

        if isinstance(source, str):
            self.http_client = HttpRangeClient(source)
        elif isinstance(source, HttpRangeClient):
            self.http_client = source
        else:
            raise ValueError("Unknown source")

    async def get_range_async(
        self, start: int, length: int, min_req_length: int, purpose: str
    ) -> bytes:
        self.bytes_ever_used += length

        start_i = start - self.head
        end_i = start_i + length
        if start_i >= 0 and end_i <= len(self.buffer):
            return self.buffer[start_i:end_i]

        length_to_fetch = max(length, min_req_length)

        self.bytes_ever_fetched += length_to_fetch
        self.buffer = await self.http_client.get_range_async(
            start, length_to_fetch, purpose
        )
        self.head = start

        return self.buffer[:length]

    def get_range(
        self, start: int, length: int, min_req_length: int, purpose: str
    ) -> bytes:
        self.bytes_ever_used += length

        start_i = start - self.head
        end_i = start_i + length
        if start_i >= 0 and end_i <= len(self.buffer):
            return self.buffer[start_i:end_i]

        length_to_fetch = max(length, min_req_length)

        self.bytes_ever_fetched += length_to_fetch
        self.buffer = self.http_client.get_range(start, length_to_fetch, purpose)
        self.head = start

        return self.buffer[:length]

    def log_usage(self, purpose: str) -> None:
        category = purpose.split(" ")[0]
        used = self.bytes_ever_used
        requested = self.bytes_ever_fetched
        efficiency = f"{(100.0 * used / requested):.2f}"

        logger.info(
            f"{category} bytes used/requested: {used} / {requested} = {efficiency}%"
        )


class HttpRangeClient:
    def __init__(self, url: str):
        self.url = url
        self.requests_ever_made = 0
        self.bytes_ever_requested = 0

    async def get_range_async(self, begin: int, length: int, purpose: str) -> bytes:
        # TODO: Use aiohttp
        # raise NotImplementedError()
        self.requests_ever_made += 1
        self.bytes_ever_requested += length

        range_header = f"bytes={begin}-{begin + length - 1}"
        headers = {
            "Range": range_header,
        }

        req = urllib.request.Request(self.url, headers=headers)
        response = urllib.request.urlopen(req)

        return response.read()

    def get_range(self, begin: int, length: int, purpose: str) -> bytes:
        self.requests_ever_made += 1
        self.bytes_ever_requested += length

        range_header = f"bytes={begin}-{begin + length - 1}"
        headers = {
            "Range": range_header,
        }

        req = urllib.request.Request(self.url, headers=headers)
        response = urllib.request.urlopen(req)

        return response.read()



from flatgeobuf.geojson.reader import HTTPReader  # noqa: F401
from flatgeobuf.geojson.reader import Reader  # noqa: F401
from flatgeobuf.geojson.reader import load  # noqa: F401
from flatgeobuf.geojson.reader import load_http_async  # noqa: F401
from flatgeobuf.geojson.reader import load_http  # noqa: F401



from __future__ import annotations

from typing import List, cast

from geojson import (
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)

from flatgeobuf.FlatGeobuf.Geometry import Geometry
from flatgeobuf.FlatGeobuf.GeometryType import GeometryType
from flatgeobuf.generic.geometry import (
    ParsedGeometry,
    flat,
    pair_flat_coordinates,
    to_geometry_type,
)


def parse_geometry(
    geometry: (
        Point | MultiPoint | LineString | MultiLineString | Polygon | MultiPolygon
    ),
) -> ParsedGeometry:
    cs = geometry.coordinates
    xy = []
    z = []
    ends = None
    parts = None
    type = to_geometry_type(geometry.type)
    end = 0
    if geometry.type == "Point":
        flat(cs, xy, z)
    elif geometry.type in ["MultiPoint", "LineString"]:
        flat(cs, xy, z)
    elif geometry.type in ["MultiLineString", "Polygon"]:
        css = cs
        flat(css, xy, z)
        if len(css) > 1:
            ends = [end + len(c) for c in css]
    elif geometry.type == "MultiPolygon":
        csss = cs
        geometries = [
            cast(Polygon, {"type": "Polygon", "coordinates": coordinates})
            for coordinates in csss
        ]
        parts = [parse_geometry(geometry) for geometry in geometries]

    parsed_geometry: ParsedGeometry = {
        "xy": xy,
        "z": z if len(z) > 0 else None,
        "ends": ends,
        "type": type,
        "parts": parts,
    }
    return parsed_geometry


def parse_gc(geometry: GeometryCollection) -> ParsedGeometry:
    type = to_geometry_type(geometry.type)
    parts = []
    for g in geometry.geometries:
        if g.type == "GeometryCollection":
            parts.append(parse_gc(g))
        else:
            parts.append(parse_geometry(g))

    parsed_geometry: ParsedGeometry = {
        "type": type,
        "parts": parts,
    }
    return parsed_geometry


def extract_parts(xy: List[float], z: List[float] | None, ends: List[int] | None):
    if ends is None or len(ends) == 0:
        return [pair_flat_coordinates(xy, z)]

    s = 0
    xy_slices = [xy[s : (s := e << 1)] for e in ends]

    z_slices = None
    if z:
        s = 0
        z_slices = [z[s : (s := e)] for e in ends]

    return [
        pair_flat_coordinates(xy, z_slices[i] if z_slices else None)
        for i, xy in enumerate(xy_slices)
    ]


def to_geojson_coordinates(geometry: Geometry, type: GeometryType):
    xy = geometry.XyAsNumpy()
    if isinstance(xy, int) and xy == 0:
        raise ValueError("Geometry has no xy coordinates")
    else:
        xy = xy.tolist()

    z = geometry.ZAsNumpy()
    z = z.tolist() if not (isinstance(z, int) and z == 0) else None

    if type == GeometryType.Point:
        a = list(xy)
        if z:
            a.append(z[0])
        return a
    elif type in [GeometryType.MultiPoint, GeometryType.LineString]:
        return pair_flat_coordinates(xy, z)
    elif type in [GeometryType.MultiLineString, GeometryType.Polygon]:
        ends = geometry.EndsAsNumpy()
        ends = ends.tolist() if not (isinstance(ends, int) and ends == 0) else None
        return extract_parts(xy, z, ends)


def from_geometry(geometry: Geometry, header_type: GeometryType):
    type = header_type

    if type == GeometryType.Unknown:
        type = geometry.Type()

    if type == GeometryType.GeometryCollection:
        geometries = []

        for i in range(geometry.PartsLength()):
            part = geometry.Parts(i)
            part_type = part.Type()
            geometries.append(from_geometry(part, part_type))

        return GeometryCollection(geometries)

    elif type == GeometryType.MultiPolygon:
        geometries = []

        for i in range(geometry.PartsLength()):
            geometries.append(from_geometry(geometry.Parts(i), GeometryType.Polygon))

        return MultiPolygon([g["coordinates"] for g in geometries])

    coordinates = to_geojson_coordinates(geometry, type)

    # NOTE: Modified
    geometry_type = next(filter(lambda x: x[1] == type, GeometryType.__dict__.items()))[
        0
    ]

    return {"type": geometry_type, "coordinates": coordinates}
